/* blssl/SSLGlobal.c */
/* ====================================================================
 * Copyright (C) 2002 Jung-Ho Cha (hinunbi@users.kldp.net).  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. Neither the name of the "Barunmo" nor the names of its 
 *    contributors may not be used to endorse or promote products 
 *    derived from this software without specific prior written 
 *    permission. 
 *    For written permission, please contact hinunbi@users.kldp.net.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
 * THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes software developed by the OpenSSL Project
 * for use in the OpenSSL Toolkit. (http://www.openssl.org/)
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */


#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <openssl/conf.h>
#include <openssl/err.h>

#include "com_brm_net_ssl_SSLGlobal.h"
#include "brm_ssl.h"
#include "brm_util.h"

/* defined in th-lock.c */
void CRYPTO_thread_setup(void);
void CRYPTO_thread_cleanup(void);


/*
 * Define the SSL Protocol options from mod_ssl
 */
#define SSL_PROTOCOL_NONE  (0)
#define SSL_PROTOCOL_SSLV2 (1<<0)
#define SSL_PROTOCOL_SSLV3 (1<<1)
#define SSL_PROTOCOL_TLSV1 (1<<2)
#define SSL_PROTOCOL_ALL   \
		(SSL_PROTOCOL_SSLV2|SSL_PROTOCOL_SSLV3|SSL_PROTOCOL_TLSV1)


/*
 * Class:     com_brm_net_ssl_SSLGlobal
 * Method:    nativeNew
 * Signature: (Lcom/brm/net/ssl/SSLConf;Ljava/lang/String;)J
 */
JNIEXPORT jlong JNICALL Java_com_brm_net_ssl_SSLGlobal_nativeNew
  (JNIEnv *env, 
  jobject obj, 
  jobject conf, 
  jstring sslCertificateKeyPasswd)
{
	int ret = 0;
	char *fname;
	jboolean isCopy = JNI_FALSE;
	BIO *bio = NULL;
	char *p;
	BRM_SSL_CTX *bctx;
	SSL_CTX *ctx = NULL;
	char *value;
	int ivalue;
	int protocol, verify;
	char *passwd = NULL;
	X509_STORE *store;
    X509_LOOKUP *lookup;
	static int thread_setup = 0;

	SSL_load_error_strings();
	SSL_library_init();

#ifdef OPENSSL_THREADS
	if(!thread_setup) {
		CRYPTO_thread_setup();
		thread_setup = 1;
	}
#endif

	bio = BIO_new(BIO_s_mem());
	if(!bio) {
		fprintf(stderr, "brm : unable to allocate the bio io\n");
		raise_exception(env, NULL);
		return 0;
	}
	bctx = OPENSSL_malloc(sizeof(BRM_SSL_CTX));
	if(!bctx) {
		ERR_print_errors(bio);
		BIO_printf(bio, 
			"brm : unable to allocate the memory size %d bytes\n", 
			sizeof(BRM_SSL_CTX));
		goto err;
	}
	memset(bctx, 0, sizeof(BRM_SSL_CTX));

	/* SSLMessageTrace ----------------------------------------------------- */
	fname = "SSLMessageTrace"; 
	get_jstr(env, conf, fname, &value);
	if(value) {
		p = value;
		while(*p) *p++ = (char)toupper(*p);
		if(strstr(value, "ON")) bctx->SSLMessageTrace = ~0;
		OPENSSL_free(value);
	}

	/* SSLHandShakeTimeout ------------------------------------------------- */
	fname = "SSLHandShakeTimeout"; 
	get_jint(env, conf, fname, &ivalue);
	bctx->SSLHandShakeTimeout = ivalue;

	/* SSLProtocol --------------------------------------------------------- */
	fname = "SSLProtocol";
	get_jstr(env, conf, fname, &value);
	if(value) {
		protocol = 0;
		chk_protocol_option(value, "ALL", SSL_PROTOCOL_ALL, &protocol);
		ret = chk_protocol_option(
				value, "SSLV2", SSL_PROTOCOL_SSLV2, &protocol);
		ret = chk_protocol_option(
				value, "SSLV3", SSL_PROTOCOL_SSLV3, &protocol);
		ret = chk_protocol_option(
				value, "TLSv1", SSL_PROTOCOL_TLSV1, &protocol);
		OPENSSL_free(value); 
		
		if(protocol == SSL_PROTOCOL_SSLV2) 
			ctx = SSL_CTX_new(SSLv2_method());
		else if(protocol == SSL_PROTOCOL_SSLV3) 
			ctx = SSL_CTX_new(SSLv3_method());
		else
			ctx = SSL_CTX_new(SSLv23_method());

		if(!ctx) {
			ERR_print_errors(bio);
			BIO_printf(bio, "brm : unable to allocate SSL_CTX\n");
			goto err;
		}

		SSL_CTX_set_options(ctx, SSL_OP_ALL);
		if (!(protocol & SSL_PROTOCOL_SSLV2))
			SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
		if (!(protocol & SSL_PROTOCOL_SSLV3))
			SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv3);
		if (!(protocol & SSL_PROTOCOL_TLSV1))
			SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1);
	}

	/* SSLSessionCacheTimeout ---------------------------------------------- */
	fname = "SSLSessionCacheTimeout"; 
	get_jint(env, conf, fname, &ivalue);
	bctx->SSLSessionCacheTimeout = ivalue;
	if(ivalue > 0) 
		SSL_CTX_set_timeout(ctx, ivalue);

	/* SSLSessionCache ----------------------------------------------------- *
	fname = "SSLSessionCache"; 
	get_jstr(env, conf, fname, &value);
	bctx->SSLSessionCache = value; */

	/* SSLCipherSuite ----------------------------------------------------- */
	fname = "SSLCipherSuite"; 
	get_jstr(env, conf, fname, &value);
    if(value) {
        ret = SSL_CTX_set_cipher_list(ctx, value); 
		OPENSSL_free(value);
		if(!ret) {
			ERR_print_errors(bio);
			BIO_printf(bio, "brm : unable to set SSLCipherSuite\n");
			goto err;
		}
    }

    /* SSLCACertificatePath ------------------------------------------------ */
	fname = "SSLCACertificatePath"; 
	get_jstr(env, conf, fname, &value);
    if(value) {
        ret = SSL_CTX_load_verify_locations(ctx, NULL, value);
		OPENSSL_free(value);
		if(!ret) {
			ERR_print_errors(bio);
			BIO_printf(bio, "brm : unable to load SSLCACertificatePath\n");
			goto err;
        }
    }

    /* SSLCARevocationPath ------------------------------------------------- */
	fname = "SSLCARevocationPath"; 
	get_jstr(env, conf, fname, &value);
    if(value) {
		store = X509_STORE_new();
		if(!store) {
			OPENSSL_free(value);
			ERR_print_errors(bio);
			BIO_printf(bio, "brm : unable to new X509_STORE "
							"when inintialzing SSLCARevocationPath\n");
			goto err;
	   }
       lookup = X509_STORE_add_lookup(store, X509_LOOKUP_hash_dir());
	   if(!lookup) {
            X509_STORE_free(store);
            OPENSSL_free(value);
			ERR_print_errors(bio);
			BIO_printf(bio, "brm : unable to set X509_STORE_add_lookup "
							"when inintialzing SSLCARevocationPath\n");
			goto err;
        }
        X509_LOOKUP_add_dir(lookup, value, X509_FILETYPE_PEM);
    	bctx->RevocationStore = store;
    	OPENSSL_free(value);
    }
	/* SSLCARevocationAutoUpdate ------------------------------------------- */
	fname = "SSLCARevocationAutoUpdate"; 
	get_jstr(env, conf, fname, &value);
	if(value) {
		p = value;
		while(*p) *p++ = (char)toupper(*p);
		if(strstr(value, "ON")) bctx->SSLCARevocationAutoUpdate = ~0;
		OPENSSL_free(value);
	}

    /* SSLCertificateFile --------------------------------------------------- */
	fname = "SSLCertificateFile"; 
	get_jstr(env, conf, fname, &value);
	if(value) {
		if(SSL_CTX_use_certificate_file(
			ctx, value,	SSL_FILETYPE_PEM) <= 0) {
			if(SSL_CTX_use_certificate_file(
				ctx, value,	SSL_FILETYPE_ASN1) <= 0) {
				OPENSSL_free(value);
				ERR_print_errors(bio);
				BIO_printf(bio, 
					"brm : unable to initalize SSLCertificateFile\n");
				goto err;
			}
		}
		OPENSSL_free(value);
	}

    /* SSLCertificateFile -------------------------------------------------- */
	fname = "SSLCertificateKeyFile"; 
	get_jstr(env, conf, fname, &value);
	if(value) {
		if(sslCertificateKeyPasswd) {
			passwd = (char*)(*env)->GetStringUTFChars(env, 
									sslCertificateKeyPasswd, &isCopy);
			if(!passwd) {
				OPENSSL_free(value);
				BIO_printf(bio, "brm : sslCertificateKeyPasswd read error\n");
				goto err;
			} 
		}
		ret = SSL_CTX_use_PrivateKey_file_ex(
				ctx, value, SSL_FILETYPE_PEM, passwd);
		if( ret <= 0) {
			if(SSL_CTX_use_PrivateKey_file_ex(
				ctx, value, SSL_FILETYPE_ASN1, passwd) <= 0) {
				if(isCopy == JNI_TRUE) 
					(*env)->ReleaseStringUTFChars(
							env, sslCertificateKeyPasswd, passwd);
				OPENSSL_free(value);
				ERR_print_errors(bio);
				BIO_printf(bio, 
					"brm : unable to initalize SSLCertificateKeyFile\n");
				goto err;
			}
		}
		if(isCopy == JNI_TRUE) 
			(*env)->ReleaseStringUTFChars(
						env, sslCertificateKeyPasswd, passwd);
		OPENSSL_free(value);
	}

     /* SSLVerify ---------------------------------------------------------- */
	fname = "SSLVerify"; 
	get_jstr(env, conf, fname, &value);
	if(value) {
		verify = SSL_VERIFY_NONE;
		if(strstr(value, "none")) verify |= SSL_VERIFY_NONE; 
		else if(strstr(value, "optional")) 
			verify |= SSL_VERIFY_PEER;
		else if(strstr(value, "require"))  
			verify |= SSL_VERIFY_PEER | SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
		else if(strstr(value, "optional_no_ca")) 
			verify |= SSL_VERIFY_PEER;
		SSL_CTX_set_verify(ctx, verify, cert_verify_cb);
		OPENSSL_free(value);
	}

	 /* SSLVerifyDepth ----------------------------------------------------- */
	fname = "SSLVerifyDepth";
	get_jint(env, conf, fname, &ivalue); 
	SSL_CTX_set_verify_depth(ctx, ivalue);

	/* Ʒ ɼ  ҿ CRL Բ     ִ 
	̴. 
	store = SSL_CTX_get_cert_store(ctx);
	vflags |= X509_V_FLAG_CRL_CHECK|X509_V_FLAG_CRL_CHECK_ALL; 
	X509_STORE_set_flags(store, vflags); */

	bctx->ctx = ctx;
	BIO_free(bio);
	return (jlong)bctx;
err:
	raise_exception(env, bio);
	if(bio) BIO_free(bio);
	if(bctx) BRM_SSL_CTX_Free(bctx);
	return 0;
}

/*
 * Class:     com_brm_net_ssl_SSLGlobal
 * Method:    nativeFree
 * Signature: (J)V
 */
JNIEXPORT void JNICALL Java_com_brm_net_ssl_SSLGlobal_nativeFree
  (JNIEnv *env, jobject obj, jlong in)
{
	BRM_SSL_CTX *bctx = (BRM_SSL_CTX *)in;

	if(bctx) BRM_SSL_CTX_Free(bctx);
}


